#!/usr/bin/python
## -*- utf-8 -*-

import threading
import socket
import select
import logging
log = logging.getLogger(__name__) 
from util import *

##############################################################################
## Interface

class Interface(object):
  
  def intro(self,name):
    log.debug("message: intro (unhandled)")
    
  def bye(self):
    log.debug("message: bye (unhandled)")
    
  def start_game(self):
    log.debug("message: start_game (unhandled)")

  def end_game(self):
    log.debug("message: end_game (unhandled)")
    
  def sequence(self,seq):
    log.debug("message: sequence (unhandled) %(seq)s" % vars())

  def winner(self,name):
    log.debug("message: winner (unhandled)")


##############################################################################
##

class Mode(object):
  def __init__(self,ctx):
    self._ctx = ctx
  def name(self): raise RuntimeError("no implementation")
  def start(self): raise RuntimeError("invalid mode for start")
  def close(self): raise RuntimeError("invalid mode for stop")

class Mode_Idle(Mode):
  def __init__(self,ctx):
    super(Mode_Idle,self).__init__(ctx)
  def name(self):
    return "idle"

class Mode_AsHost(Mode):
  def __init__(self,ctx):
    super(Mode_AsHost,self).__init__(ctx)
    self._athread = None
    self._dthread = None
    self._asocket = None
    self._dsocket = None
    self._dfile = None
    self._closing = False
    
  def name(self):
    return "ashost"
  
  def dfile(self):
    return self._dfile
  
  def start(self):
    self._asocket = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    self._asocket.bind((socket.gethostname(),self._ctx._port))
    self._asocket.listen(5)
    self._athread = threading.Thread(target=self.run_for_listen)
    self._athread.start()
        
  def close(self):
    self._closing = True
    self._asocket.close()
    if self._dsocket:
      self._dsocket.close()
    self._athread = None
    self._dthread = None
    self._asocket = None
    self._dsocket = None

  def run_for_data(self):
    while not self._closing:
      rr,rw,err = select.select( [self._dsocket], [], [], 1.0)
      if rr:
        line = self._dfile.readline() # WAITING
        if len(line) == 0: # close
          self._ctx._ind.bye()
          self._ctx.close()
        else:
          #log.debug("line: %(line)s" % vars())
          self._ctx.dispatch(line)
  
  def run_for_listen(self):
    port = self._ctx._port
    log.debug("waiting connections from port %(port)s" %vars())
    while not self._closing:
      rr,rw,err = select.select( [self._asocket], [], [], 1.0)
      if rr:
        (dsocket,addr) = self._asocket.accept() # WAITING
        log.debug("new connection from %(addr)s" % vars())
        assert(self._dsocket == None)
        self._dsocket = dsocket
        self._dfile = dsocket.makefile() 
        self._dthread = threading.Thread(target=self.run_for_data)
        self._dthread.start()

class Mode_AsClient(Mode):

  def __init__(self,ctx):
    super(Mode_AsClient,self).__init__(ctx)
    self._dthread = None
    self._dsocket = None
    self._dfile = None
    self._closing = False
      
  def name(self):
    return "asclient"

  def target(self,ip,port):
    self._ip = ip
    self._port = port

  def dfile(self):
    return self._dfile
  
  def start(self):
    self._dsocket = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    ip = self._ip
    port = self._port
    log.debug("connecting to %(ip)s:%(port)s" % vars())
    self._dsocket.connect((ip,port))
    self._dfile = self._dsocket.makefile()
    self._dthread = threading.Thread(target=self.run_for_data)
    self._dthread.start()
        
  def close(self):
    self._closing = True
    self._dsocket.close()
    self._dthread = None
    self._dsocket = None
  
  def run_for_data(self):
    while not self._closing:
      rr,rw,err = select.select( [self._dsocket], [], [], 1.0)
      if rr:
        line = self._dfile.readline() # WAITING
        if len(line) == 0: # close
          self._ctx._ind.bye()
          self._ctx.close()
        else:
          #log.debug("line: %(line)s" % vars())
          self._ctx.dispatch(line)

class Comm(object):
  def __init__(self,ind=None):
    self._ind = ind # indication object
    self.IDLE = Mode_Idle(self)
    self.ASHOST = Mode_AsHost(self)
    self.ASCLIENT = Mode_AsClient(self)
    self.change_mode(self.IDLE)
    self._mode = None
    self._port = 12345
  
  def set_ind(self,ind):
    self._ind = ind
  
  def mode(self):
    return self._mode
  
  def start_host(self):
    self.change_mode(self.ASHOST)
    self._mode.start()
  
  def start_client(self,ip,port=None):
    self.change_mode(self.ASCLIENT)
    if port is None:
      port = self._port
    self._mode.target(ip,port)
    self._mode.start()
  
  def close(self):
    self._mode.close()
    self.change_mode(self.IDLE)

  def change_mode(self,new_mode):
    mode_name = new_mode.name()
    log.debug("Comm: mode: %(mode_name)s" % vars())
    self._mode = new_mode

  def dispatch(self,line):
    ind = self._ind
    parts = line.split()
    msg = parts[0]
    if msg == "intro":
      ind.intro(parts[1])
    elif msg == "bye":
      ind.bye()
    elif msg == "start_game":
      ind.start_game()
    elif msg == "end_game":
      ind.end_game()
    elif msg == "sequence":
      seq = parts[1].split("-")
      ind.sequence(seq)
    elif msg == "winner":
      ind.winner(parts[1])
    else:
      raise RuntimeError("received unknown message %(msg)s" % vars())      
    
  def intro(self,name):
    dfile = self._mode.dfile()
    dfile.write("intro %(name)s\n" % vars())
    dfile.flush()
    
  def bye(self):
    dfile = self._mode.dfile()
    dfile.write("bye\n")
    dfile.flush()
    
  def start_game(self):
    dfile = self._mode.dfile()
    dfile.write("start_game\n")
    dfile.flush()
    
  def end_game(self):
    dfile = self._mode.dfile()
    dfile.write("end_game\n")
    dfile.flush()
    
  def sequence(self,seq):
    dfile = self._mode.dfile()
    seqs = map(str,seq)
    seqstr = "-".join(seqs)
    dfile.write("sequence %(seqstr)s\n" % vars())
    dfile.flush()

  def winner(self,name):
    dfile = self._mode.dfile()
    dfile.write("winner %(name)s\n" %vars())
    dfile.flush()

    


##############################################################################
##

if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG)
  comm = Comm(None)
  comm.start_client("localhost")
  comm.close()
  